# -*- coding: utf-8 -*-

__author__ = "Marie Roald & Yngve Mardal Moe"

"""
This module contains most functions that only work on tensor factorisation models, not data
tensors. The module contains functions that are useful for inspecting tensor factorisation
models. For example, computing how similar two factorisations are, checking if two decompositions
are equivalent, or simply generating a dense tensor from a (possibly) labelled decomposition.
"""

__author__ = "Marie Roald & Yngve Mardal Moe"

from warnings import warn

import numpy as np
import pandas as pd
from scipy.optimize import linear_sum_assignment

from ._module_utils import (
    _handle_none_weights_cp_tensor,
    is_dataframe,
    validate_cp_tensor,
)
from ._tl_utils import (
    _handle_tensorly_backends_cp,
    _handle_tensorly_backends_dataset
)
from ._xarray_wrapper import (
    _SINGLETON,
    _handle_labelled_cp,
    _handle_labelled_dataset,
    _handle_labelled_factor_matrix,
)
from .utils import _alias_mode_axis, cp_norm, extract_singleton, normalise

__all__ = [
    "normalise_cp_tensor",
    "distribute_weights_evenly",
    "distribute_weights_in_one_mode",
    "distribute_weights",
    "cosine_similarity",
    "get_factor_matrix_permutation",
    "factor_match_score",
    "degeneracy_score",
    "get_cp_permutation",
    "permute_cp_tensor",
    "check_factor_matrix_equal",
    "check_cp_tensor_equal",
    "check_factor_matrix_close",
    "check_cp_tensors_equivalent",
    "percentage_variation",
]


@_handle_tensorly_backends_cp("cp_tensor", _SINGLETON)
@_handle_labelled_cp("cp_tensor", _SINGLETON)
def normalise_cp_tensor(cp_tensor):
    """Ensure that the all factor matrices have unit norm, and all weight is stored in the weight-vector

    Parameters
    ----------
    cp_tensor : CPTensor or tuple
        TensorLy-style CPTensor object or tuple with weights as first
        argument and a tuple of components as second argument.

    Returns
    -------
    tuple
        The scaled CP tensor.
    """
    weights, factors = cp_tensor
    if weights is None:
        weights = np.ones(factors[0].shape[1])

    weights = weights.copy()
    new_factors = []
    for factor in factors:
        norm = np.linalg.norm(factor, axis=0, keepdims=True)
        weights *= norm.ravel()

        # If a component vector is zero, then we do not want to divide by zero, and zero / 1 is equal to zero.
        norm[norm == 0] = 1
        new_factors.append(factor / norm)
    return weights, tuple(new_factors)


@_handle_tensorly_backends_cp("cp_tensor", _SINGLETON)
@_handle_labelled_cp("cp_tensor", _SINGLETON)
def distribute_weights_evenly(cp_tensor):
    """Ensure that the weight-vector consists of ones and all factor matrices have equal norm

    Parameters
    ----------
    cp_tensor : CPTensor or tuple
        TensorLy-style CPTensor object or tuple with weights as first
        argument and a tuple of components as second argument.

    Returns
    -------
    tuple
        The scaled CP tensor.
    """
    weights, factors = normalise_cp_tensor(cp_tensor)
    weights = weights ** (1 / len(factors))
    for factor in factors:
        factor[:] *= weights
    weights = np.ones_like(weights)
    return weights, factors


@_handle_tensorly_backends_cp("cp_tensor", _SINGLETON)
@_handle_labelled_cp("cp_tensor", _SINGLETON)
@_alias_mode_axis()
def distribute_weights_in_one_mode(cp_tensor, mode, axis=None):
    """Normalise all factors and multiply the weights into one mode.

    The CP tensor is scaled so all factor matrices except one have unit norm
    columns and the weight-vector contains only ones.

    Parameters
    ----------
    cp_tensor : CPTensor or tuple
        TensorLy-style CPTensor object or tuple with weights as first
        argument and a tuple of components as second argument.
    mode : int
        Which mode (axis) to store the weights in
    axis : int (optional)
        Alias for mode. If this is set, then no value is needed for mode

    Returns
    -------
    tuple
        The scaled CP tensor.
    """
    weights, factors = normalise_cp_tensor(cp_tensor)
    factors[mode][:] *= weights
    return np.ones_like(weights), factors


def distribute_weights(cp_tensor, weight_behaviour, weight_mode=0):
    """Utility to distribute the weights of a CP tensor.

    Parameters
    ----------
    cp_tensor : CPTensor or tuple
        TensorLy-style CPTensor object or tuple with weights as first
        argument and a tuple of components as second argument.
    weight_behaviour : {"ignore", "normalise", "evenly", "one_mode"} (default="normalise")
        How to handle the component weights.

         * ``"ignore"`` - Do nothing
         * ``"normalise"`` - Normalise all factor matrices
         * ``"evenly"`` - All factor matrices have equal norm
         * ``"one_mode"`` - The weight is allocated in one mode, all other factor matrices have unit norm columns.

    weight_mode : int (optional)
        Which mode to have the component weights in (only used if ``weight_behaviour="one_mode"``)

    Returns
    -------
    tuple
        The scaled CP tensor.

    See Also
    --------
    normalise_cp_tensor : Give all component vectors unit norm
    distribute_weights_evenly : Give all component vectors the same norm and set the weight-array to one.
    distribute_weights_in_one_mode : Keep all the weights in one factor matrix and set the weight-array to one.

    Raises
    ------
    ValueError
        If ``weight_behaviour`` is not one of ``"ignore"``, ``"normalise"``, ``"evenly"`` or ``"one_mode"``.
    """
    if weight_behaviour == "ignore":
        return cp_tensor
    elif weight_behaviour == "normalise":
        return normalise_cp_tensor(cp_tensor)
    elif weight_behaviour == "evenly":
        return distribute_weights_evenly(cp_tensor)
    elif weight_behaviour == "one_mode":
        return distribute_weights_in_one_mode(cp_tensor, weight_mode)
    else:
        raise ValueError("weight_behaviour must be either 'ignore', 'normalise', 'evenly', or 'one_mode'")


@_handle_tensorly_backends_dataset("factor_matrix1", None)
@_handle_tensorly_backends_dataset("factor_matrix2", None)
@_handle_labelled_factor_matrix("factor_matrix2", None)
@_handle_labelled_factor_matrix("factor_matrix1", None)
def cosine_similarity(factor_matrix1, factor_matrix2):
    r"""The average cosine similarity (Tucker congruence) with optimal column permutation.

    The cosine similarity between two vectors is computed as

    .. math::

        \cos (\mathbf{x}, \mathbf{y}) =
        \frac{\mathbf{x}^\mathsf{T}}{\|\mathbf{x}\|}\frac{\mathbf{y}}{\|\mathbf{y}\|}

    This function returns the average cosine similarity between the columns vectors of
    the two factor matrices, using the optimal column permutation.

    Parameters
    ----------
    factor_matrix1 : np.ndarray or pd.DataFrame
        First factor matrix
    factor_matrix2 : np.ndarray or pd.DataFrame
        Second factor matrix

    Returns
    -------
    float
        The average cosine similarity.
    """
    congruence = normalise(factor_matrix1).T @ normalise(factor_matrix2)
    permutation = linear_sum_assignment(-congruence)
    return extract_singleton(congruence[permutation].mean())


NO_COLUMN = slice(0, 0, 1)


def _get_linear_sum_assignment_permutation(cost_matrix, allow_smaller_rank):
    row_index, column_index = linear_sum_assignment(-cost_matrix)

    R1, R2 = cost_matrix.shape
    if R1 > R2 and not allow_smaller_rank:
        raise ValueError(
            f"Cannot permute a {R2}-column matrix against a {R1}-column matrix unless ``allow_smaller_rank=True``."
        )

    permutation = [None] * max(R1, R2)
    for row_idx, col_idx in zip(row_index, column_index):
        permutation[row_idx] = col_idx

    missing_entries = sorted(set(range(R2)) - set(permutation))
    for i, missing in enumerate(missing_entries):
        permutation[i + len(row_index)] = missing

    for i, p, in enumerate(permutation):
        if p is None:
            permutation[i] = NO_COLUMN

    return row_index, column_index, permutation


@_handle_tensorly_backends_dataset("factor_matrix1", None)
@_handle_tensorly_backends_dataset("factor_matrix2", None)
def get_factor_matrix_permutation(factor_matrix1, factor_matrix2, ignore_sign=True, allow_smaller_rank=False):
    r"""Find optimal permutation of the factor matrices

    Efficient estimation of the optimal permutation for two factor matrices.
    To find the optimal permutation, :math:`\sigma`, we solve the following
    optimisation problem:

    .. math::

        \max_\sigma \sum_{r} \frac{\left|\mathbf{a}_{r}^\mathsf{T}\hat{\mathbf{a}}_{\sigma(r)}\right|}
                                  {\|\mathbf{a}_{r}\| \|\hat{\mathbf{a}}_{\sigma(r)}\|}

    where :math:`\mathbf{a}_r` is the :math:`r`-th component vector for the
    first factor matrix and :math:`\hat{\mathbf{a}}_{\sigma(r)}` is :math:`r`-th
    component vector of the second factor matrix after permuting the columns.

    Parameters
    ----------
    factor_matrix1 : np.ndarray or pd.DataFrame
        First factor matrix
    factor_matrix2 : np.ndarray or pd.DataFrame
        Second factor matrix
    ignore_sign : bool
        Whether to take the absolute value of the inner products before
        computing the permutation. This is usually done because of the sign
        indeterminacy of component models.
    allow_smaller_rank : bool (default=False)
        If ``True``, then the function can align a smaller matrix onto a larger one. Missing
        columns are aligned with ``tlviz.factor_tools.NO_COLUMN`` (a slice that slices nothing).

    Returns
    -------
    permutation : list[int | slice]
        List of ints used to permute ``factor_matrix2`` so its columns optimally align with ``factor_matrix1``.
        If the ``factor_matrix1`` has a column with no corresponding column in ``factor_matrix2`` (i.e. there
        are fewer columns in ``factor_matrix2`` than in ``factor_matrix1``), then
        ``tlviz.factor_tools.NO_COLUMN`` (a slice that slices nothing) is used to indicate missing columns.

    Raises
    ------
    ValueError
        If ``allow_smaller_rank=False`` and ``factor_matrix2`` has fewer columns than ``factor_matrix1``.
    """
    congruence_product = normalise(factor_matrix1).T @ normalise(factor_matrix2)
    if ignore_sign:
        congruence_product = np.abs(congruence_product)

    return _get_linear_sum_assignment_permutation(congruence_product, allow_smaller_rank=allow_smaller_rank)[-1]


@_handle_tensorly_backends_cp("cp_tensor1", None)
@_handle_tensorly_backends_cp("cp_tensor2", None)
def factor_match_score(
    cp_tensor1,
    cp_tensor2,
    consider_weights=True,
    skip_mode=None,
    return_permutation=False,
    absolute_value=True,
    allow_smaller_rank=False,
):
    r"""Compute the factor match score between ``cp_tensor1`` and ``cp_tensor2``.

    The factor match score is used to measure the similarity between two
    sets of components. There are many definitions of the FMS, but one
    common definition for third order tensors is given by:

    .. math::

        \sum_{r=1}^R \frac{\mathbf{a}_r^T \hat{\mathbf{a}}_r}{\|\mathbf{a}_r^T\| \|\hat{\mathbf{a}}_r\|}
                     \frac{\mathbf{b}_r^T \hat{\mathbf{b}}_r}{\|\mathbf{b}_r^T\| \|\hat{\mathbf{b}}_r\|}
                     \frac{\mathbf{c}_r^T \hat{\mathbf{c}}_r}{\|\mathbf{c}_r^T\| \|\hat{\mathbf{c}}_r\|},

    where :math:`\mathbf{a}, \mathbf{b}` and :math:`\mathbf{c}` are the component vectors for
    one of the decompositions and :math:`\hat{\mathbf{a}}, \hat{\mathbf{b}}` and :math:`\hat{\mathbf{c}}`
    are the component vectors for the other decomposition. Often, the absolute value of the inner
    products is used instead of just the inner products (i.e. :math:`|\mathbf{a}_r^T \hat{\mathbf{a}}_r|`).

    The above definition does not take the norm of the component vectors into account.
    However, sometimes, we also wish to compare their norm. In that case, set the
    ``consider_weights`` argument to ``True`` to compute

    .. math::

        \sum_{r=1}^R \left(1 - \frac{w_r \hat{w}_r}{\max\left( w_r \hat{w}_r \right)}\right)
                     \frac{\mathbf{a}_r^T \hat{\mathbf{a}}_r}{\|\mathbf{a}_r^T\|\|\hat{\mathbf{a}}_r\|}
                     \frac{\mathbf{b}_r^T \hat{\mathbf{b}}_r}{\|\mathbf{b}_r^T\|\|\hat{\mathbf{b}}_r\|}
                     \frac{\mathbf{c}_r^T \hat{\mathbf{c}}_r}{\|\mathbf{c}_r^T\|\|\hat{\mathbf{c}}_r\|}

    instead, where :math:`w_r = \|\mathbf{a}_r\| \|\mathbf{b}_r\| \|\mathbf{c}_r\|` and
    :math:`\hat{w}_r = \|\hat{\mathbf{a}}_r\| \|\hat{\mathbf{b}}_r\| \|\hat{\mathbf{c}}_r\|`.

    For both definitions above, there is a permutation determinacy. Two equivalent decompositions
    can have the same component vectors, but in a different order. To resolve this determinacy,
    we use linear sum assignment solver available in SciPy :cite:p:`crouse2016implementing` to
    efficiently find the optimal permutation.

    Parameters
    ----------
    cp_tensor1 : CPTensor or tuple
        TensorLy-style CPTensor object or tuple with weights as first
        argument and a tuple of components as second argument
    cp_tensor2 : CPTensor or tuple
        TensorLy-style CPTensor object or tuple with weights as first
        argument and a tuple of components as second argument
    consider_weights : bool (default=True)
        If False, then the weight-penalty is used (second equation above).
    skip_mode : int or None (default=None)
        Which mode to skip when computing the FMS. Useful if cross validation
        or split-half analysis is used.
    return_permutation : bool (default=False)
        Whether or not to return the optimal permutation of the factors
    absolute_value : bool (default=True)
        If True, then only magnitude of the congruence is considered, not the
        sign.
    allow_smaller_rank : bool (default=False)
        Only relevant if ``return_permutation=True``. If ``True``, then ``cp_tensor2``
        can have fewer components than ``cp_tensor2``. Missing components are aligned
        with ``tlviz.factor_tools.tlviz.factor_tools.NO_COLUMN`` (a slice that slices nothing).

    Returns
    -------
    fms : float
        The factor match score
    permutation : list[int | object] (only if return_permutation=True)
        List of ints used to permute ``cp_tensor2`` so its components optimally align with ``cp_tensor1``.
        If the ``cp_tensor1`` has a component with no corresponding component in ``cp_tensor2`` (i.e. there
        are fewer components in ``cp_tensor2`` than in ``cp_tensor1``), then
        ``tlviz.factor_tools.NO_COLUMN`` (a slice that slices nothing) is used to indicate missing components.

    Raises
    ------
    ValueError
        If ``allow_smaller_rank=False`` and ``cp_tensor2`` has fewer components than ``cp_tensor1``.

    Examples
    --------
    >>> import numpy as np
    >>> from tlviz.factor_tools import factor_match_score
    >>> from tensorly.decomposition import parafac
    >>> from tensorly.random import random_cp
    >>> # Construct random cp tensor with TensorLy
    >>> cp_tensor = random_cp(shape=(4,5,6), rank=3, random_state=42)
    >>> X = cp_tensor.to_tensor()
    >>> # Add noise
    >>> X_noisy = X + 0.05*np.random.RandomState(0).standard_normal(size=X.shape)
    >>> # Decompose with TensorLy and compute FMS
    >>> estimated_cp_tensor = parafac(X_noisy, rank=3, random_state=42)
    >>> fms_with_weight_penalty = factor_match_score(cp_tensor, estimated_cp_tensor, consider_weights=True)
    >>> print(f"Factor match score (with weight penalty): {fms_with_weight_penalty:.2f}")
    Factor match score (with weight penalty): 0.95
    >>> fms_without_weight_penalty = factor_match_score(cp_tensor, estimated_cp_tensor, consider_weights=False)
    >>> print(f"Factor match score (without weight penalty): {fms_without_weight_penalty:.2f}")
    Factor match score (without weight penalty): 0.99
    """
    if skip_mode is not None and consider_weights:
        warn(
            "Cannot consider weights when a mode is skipped due to the scaling indeterminacy of PARAFAC models."
            + " consider_weights will therefore be set to False. To supress this warning, specify"
            + " consider_weights=False when calling factor_match_score with skip_mode not equal to None."
        )
        consider_weights = False

    # Extract weights and components from decomposition
    weights1, factors1 = normalise_cp_tensor(cp_tensor1)
    weights2, factors2 = normalise_cp_tensor(cp_tensor2)

    congruence_product = 1
    for i, (factor1, factor2) in enumerate(zip(factors1, factors2)):
        if hasattr(factor1, "values"):
            factor1 = factor1.values
        if hasattr(factor2, "values"):
            factor2 = factor2.values

        if i == skip_mode:
            continue
        congruence_product *= factor1.T @ factor2

    if consider_weights:
        congruence_product *= 1 - np.abs(weights1[:, np.newaxis] - weights2[np.newaxis, :]) / np.maximum(
            weights1[:, np.newaxis], weights2[np.newaxis, :]
        )

    if absolute_value:
        congruence_product = np.abs(congruence_product)

    # If permutation is not returned, then smaller rank is OK
    allow_smaller_rank = allow_smaller_rank or not return_permutation
    row_index, column_index, permutation = _get_linear_sum_assignment_permutation(
        congruence_product, allow_smaller_rank=allow_smaller_rank
    )
    congruence_product = congruence_product[row_index, column_index]

    if not return_permutation:
        return congruence_product.mean()

    return congruence_product.mean(), permutation


@_handle_tensorly_backends_cp("cp_tensor", None)
def degeneracy_score(cp_tensor):
    r"""Compute the degeneracy score for a given decomposition.

    PARAFAC models can be degenerate, which is a sign that we should
    be careful before interpreting that model. For a third order tensor,
    this generally manifests in a triple cosine of two components that
    approach -1. That is

    .. math::

        \cos(\mathbf{a}_{r}, \mathbf{a}_{s})
        \cos(\mathbf{b}_{r}, \mathbf{b}_{s})
        \cos(\mathbf{c}_{r}, \mathbf{c}_{s})
        \approx -1

    for some :math:`r \neq s`, where :math:`\mathbf{A}, \mathbf{B}`
    and :math:`\mathbf{C}` are factor matrices and

    .. math::

        \cos(\mathbf{x}, \mathbf{y}) =
        \frac{\mathbf{x}^\mathsf{T} \mathbf{y}}{\|\mathbf{x}\| \|\mathbf{y}\|}.

    Furthermore, the magnitude of the degenerate components are unbounded and
    will approach infinity as the number of iterations increase.

    Degenerate solutions typically signify that the decomposition is unreliable,
    and one should take care before interpreting the components. Degeneracy
    can, in fact, be a sign that the PARAFAC problem is ill-posed. There are certain
    tensors where there are no solutions to the least squares problem to needed to fit
    PARAFAC models. And in those cases, the "optimal" but unobtainable PARAFAC
    decomposition will have component vectors with infinite norm that point in
    opposite directions :cite:p:`krijnen2008non`.

    There are several strategies to avoid degenerate solutions:

     * Fitting models with more random initialisations
     * Decreasing the convergence tolerance or increasing the number of iterations
     * Imposing non-negativity constraints in all modes
     * Imposing orthogonality constraints in at least one mode
     * Changing the number of components

    Both non-negativity constraints and orthogonality constraints will
    remove the potential ill-posedness of the CP model. We can, in fact,
    not obtain degenerate solutions when we impose such constriants
    :cite:p:`krijnen2008non`

    To measure degeneracy, we compute the degeneracy score, which is the
    minimum triple cosine (for a third-order tensor). A score close to
    -1 signifies a degenerate solution. A score of -0.85 is an indication
    of a troublesome model :cite:p:`krijnen1993analysis` (as cited in
    :cite:p:`bro1997parafac`).

    For more information about degeneracy for component models see
    :cite:p:`zijlstra2002degenerate` and :cite:p:`bro1997parafac`.


    .. note::

        There are other kinds of degeneracies too. For example three-component
        degeneracies, which manifests as two components of increasing magnitude
        and one other component equal to the negative sum of the former
        two :cite:p:`paatero2000construction,stegeman2006degeneracy`. However, it
        is the two-component degeneracy that is most commonly discussed in the
        litterature :cite:p:`bro1997parafac,zijlstra2002degenerate,krijnen2008non`.
        Still, if three or more components display weights that have a much higher
        magnitude than the data, there is a reason to be concerned.

    Parameters
    ----------
    cp_tensor : CPTensor or tuple
        TensorLy-style CPTensor object or tuple with weights as first
        argument and a tuple of components as second argument.

    Returns
    -------
    degeneracy_score : float
        Degeneracy score, between 1 and -1. A score close to -1 signifies
        a degenerate solution. A score of -0.85 is an indication of a
        troublesome model :cite:p:`krijnen1993analysis` (as cited in
        :cite:p:`bro1997parafac`).

    Examples
    --------
    We begin by constructing a random simulated cp tensor and compute the degeneracy score

    >>> from tlviz.data import simulated_random_cp_tensor
    >>> from tlviz.factor_tools import degeneracy_score
    >>> cp_tensor = simulated_random_cp_tensor((10, 11, 12), rank=3, seed=0)[0]
    >>> print(f"Degeneracy score: {degeneracy_score(cp_tensor):.2f}")
    Degeneracy score: 0.35

    We see that (as expected) the random cp_tensor is not very degenerate. To simulate
    a tensor with two-component degeneracy, we can, for example, replace one of the
    components with a flipped copy of another component

    >>> w, (A, B, C) = cp_tensor
    >>> A[:,1] = -A[:, 0]
    >>> B[:,1] = -B[:, 0]
    >>> C[:,1] = -C[:, 0]
    >>> print(f"Degeneracy score: {degeneracy_score(cp_tensor):.2f}")
    Degeneracy score: -1.00

    We see that this modified cp_tensor is degenerate.
    """
    # TODOC: There may be some more relevant cites in Paatero 2000
    weights, factors = cp_tensor
    rank = factors[0].shape[1]
    tucker_congruence_scores = np.ones(shape=(rank, rank))

    for factor in factors:
        tucker_congruence_scores *= normalise(factor).T @ normalise(factor)

    return np.asarray(tucker_congruence_scores).min()


def _permute_cp_tensor(cp_tensor, permutation):
    """Internal function, does not handle labelled cp tensors. Use ``permute_cp_tensor`` instead.
    """
    weights, factors = cp_tensor

    if weights is not None:
        new_weights = np.zeros(len(permutation))
        for i, p in enumerate(permutation):
            if p == NO_COLUMN:
                new_weights[i] = np.nan
            else:
                new_weights[i] = weights[p]
    else:
        new_weights = None

    new_factors = [None] * len(factors)
    for mode, factor in enumerate(factors):
        new_factor = np.zeros((factor.shape[0], len(permutation)))
        for i, p in enumerate(permutation):
            if p == NO_COLUMN:
                new_factor[:, i] = np.nan
            else:
                new_factor[:, i] = factor[:, p]

        new_factors[mode] = new_factor

    return new_weights, new_factors


@_handle_tensorly_backends_cp("cp_tensor", None)
@_handle_labelled_cp("cp_tensor", None)
def get_cp_permutation(cp_tensor, reference_cp_tensor=None, consider_weights=True, allow_smaller_rank=False):
    """Find the optimal permutation between two CP tensors.

    This function supports two ways of finding the permutation of a CP tensor: Aligning the components
    with those of a reference CP tensor (if ``reference_cp_tensor`` is not ``None``), or finding the
    permutation so the components are in descending order with respect to their explained variation
    (if both ``reference_cp_tensor`` and ``permutation`` is ``None``).

    This function uses the factor match score to compute the optimal permutation between
    two CP tensors. This is useful for comparison purposes, as CP two identical CP tensors
    may have permuted columns.

    Parameters
    ----------
    cp_tensor : CPTensor or tuple
        TensorLy-style CPTensor object or tuple with weights as first
        argument and a tuple of components as second argument.
    reference_cp_tensor : CPTensor or tuple (optional)
        TensorLy-style CPTensor object or tuple with weights as first
        argument and a tuple of components as second argument. The tensor
        that ``cp_tensor`` is aligned with. Either this or the ``permutation``
        argument must be passed, not both.
    consider_weights : bool
        Whether to consider the factor weights when the factor match score is computed.

    Returns
    -------
    tuple
        The permutation to use when permuting ``cp_tensor``.
    """
    if reference_cp_tensor is not None:
        fms, permutation = factor_match_score(
            reference_cp_tensor,
            cp_tensor,
            consider_weights=consider_weights,
            return_permutation=True,
            allow_smaller_rank=allow_smaller_rank,
        )
        rank = cp_tensor[1][0].shape[1]
        target_rank = reference_cp_tensor[1][0].shape[1]

        if rank > target_rank:  # There are more components in the tensor to permute than the reference tensor
            remaining_indices = sorted(set(range(rank)) - set(permutation))
            permutation = list(permutation) + remaining_indices
    else:
        variation = percentage_variation(cp_tensor, method="model")
        permutation = sorted(range(len(variation)), key=lambda i: -variation[i])

    return permutation


@_handle_tensorly_backends_cp("cp_tensor", _SINGLETON)
@_handle_labelled_cp("cp_tensor", _SINGLETON, preserve_columns=False)
def permute_cp_tensor(
    cp_tensor, permutation=None, reference_cp_tensor=None, consider_weights=True, allow_smaller_rank=False
):
    """Permute the CP tensor

    This function supports three ways of permuting a CP tensor: Aligning the components
    with those of a reference CP tensor (if ``reference_cp_tensor`` is not ``None``),
    permuting the components according to a given permutation (if ``permutation`` is not ``None``)
    or so the components are in descending order with respect to their explained variation
    (if both ``reference_cp_tensor`` and ``permutation`` is ``None``).

    This function uses the factor match score to compute the optimal permutation between
    two CP tensors. This is useful for comparison purposes, as CP two identical CP tensors
    may have permuted columns.

    Parameters
    ----------
    cp_tensor : CPTensor or tuple
        TensorLy-style CPTensor object or tuple with weights as first
        argument and a tuple of components as second argument.
    permutation : tuple (optional)
        Tuple with the column permutations. Either this or the ``reference_cp_tensor``
        argument must be passed, not both.
    reference_cp_tensor : CPTensor or tuple (optional)
        TensorLy-style CPTensor object or tuple with weights as first
        argument and a tuple of components as second argument. The tensor
        that ``cp_tensor`` is aligned with. Either this or the ``permutation``
        argument must be passed, not both.
    consider_weights : bool
        Whether to consider the factor weights when the factor match score is computed.

    Returns
    -------
    tuple
        Tuple representing ``cp_tensor`` optimally permuted.

    Raises
    ------
    ValueError
        If neither ``permutation`` nor ``reference_cp_tensor`` is provided
    ValueError
        If both ``permutation`` and ``reference_cp_tensor`` is provided
    """
    if permutation is not None and reference_cp_tensor is not None:
        raise ValueError("Must either provide a permutation, a reference CP tensor or neither. Both is provided")

    if permutation is None:
        permutation = get_cp_permutation(
            cp_tensor=cp_tensor,
            reference_cp_tensor=reference_cp_tensor,
            consider_weights=consider_weights,
            allow_smaller_rank=allow_smaller_rank,
        )

    return _permute_cp_tensor(cp_tensor, permutation)


@_handle_tensorly_backends_dataset("factor_matrix1", None)
@_handle_tensorly_backends_dataset("factor_matrix2", None)
def check_factor_matrix_equal(factor_matrix1, factor_matrix2, ignore_labels=False):
    """Check that all entries in a factor matrix are close, if labelled, then label equality is also checked.

    This function is similar to ``numpy.allclose``, but works on both labelled and unlabelled factor
    matrices. If the factor matrices are labelled, then the DataFrame index and columns are also
    compared (unless ``ignore_labels=True``).

    Parameters
    ----------
    factor_matrix1 : numpy.ndarray or pandas.DataFrame
        Labelled or unlabelled factor matrix
    cp_tensor2 : CPTensor or tuple
        Labelled or unlabelled factor matrix
    rtol : float
        Relative tolerance (see ``numpy.allclose``)
    atol : float
        Absolute tolerance (see ``numpy.allclose``)
    ignore_labels : bool
        If True, then labels (i.e. DataFrame column names and indices) can differ.

    Returns
    -------
    bool
        Whether the decompositions are equivalent.

    Examples
    --------
    ``check_factor_matrix_equal`` checks if two factor matrices are exactly the same.

    >>> from tlviz.data import simulated_random_cp_tensor
    >>> import numpy as np
    >>> A = np.arange(6).reshape(3, 2).astype(float)
    >>> B = A.copy()
    >>> check_factor_matrix_equal(A, B)
    True

    If they are only the same up to round off errors, then this function returns ``False``

    >>> check_factor_matrix_equal(A, B + 1e-10)
    False

    If we make only one of them into a DataFrame, then the factor matrices are not equal

    >>> import pandas as pd
    >>> A_labelled = pd.DataFrame(A)
    >>> check_factor_matrix_equal(A_labelled, B)
    False
    >>> check_factor_matrix_equal(B, A_labelled)
    False

    If we turn B into a DataFrame too, it passes again

    >>> B_labelled = pd.DataFrame(A)
    >>> check_factor_matrix_equal(A_labelled, B_labelled)
    True

    The index is checked for equality, so if we change the index of ``B_labelled``, then
    the factor matrices are not equal

    >>> B_labelled.index += 1
    >>> check_factor_matrix_equal(A_labelled, B_labelled)
    False

    However, we can disable checking the labels by using the ``ignore_labels`` argument

    >>> check_factor_matrix_equal(A_labelled, B_labelled, ignore_labels=True)
    True
    """
    if is_dataframe(factor_matrix1) != is_dataframe(factor_matrix2) and not ignore_labels:
        return False
    if ignore_labels and is_dataframe(factor_matrix1):
        factor_matrix1 = factor_matrix1.values
    if ignore_labels and is_dataframe(factor_matrix2):
        factor_matrix2 = factor_matrix2.values

    if is_dataframe(factor_matrix1):
        return factor_matrix1.equals(factor_matrix2)

    return np.array_equal(factor_matrix1, factor_matrix2)


@_handle_tensorly_backends_cp("cp_tensor1", None)
@_handle_tensorly_backends_cp("cp_tensor2", None)
def check_cp_tensor_equal(cp_tensor1, cp_tensor2, ignore_labels=False):
    """Check if the factor matrices and weights are equal.

    This will check if the factor matrices and weights are exactly equal
    to one another. It will not check if the two decompositions are equivalent.
    For example, if ``cp_tensor2`` contain the same factors as ``cp_tensor1``,
    but permuted, or with the weights distributed differently between the
    modes, then this function will return False. To check for equivalence,
    use ``check_cp_tensors_equivalent``.

    Parameters
    ----------
    cp_tensor1 : CPTensor or tuple
        TensorLy-style CPTensor object or tuple with weights as first
        argument and a tuple of components as second argument
    cp_tensor2 : CPTensor or tuple
        TensorLy-style CPTensor object or tuple with weights as first
        argument and a tuple of components as second argument
    ignore_labels : bool
        If True, then labels (i.e. DataFrame column names and indices) can differ.

    Returns
    -------
    bool
        Whether the decompositions are equal.

    Examples
    --------
    ``check_cp_tensor_equal`` checks for strict equality of the factor matrices and
    weights.

    >>> from tlviz.data import simulated_random_cp_tensor
    >>> from tlviz.factor_tools import check_cp_tensor_equal
    >>> cp_tensor, dataset = simulated_random_cp_tensor((10, 20, 30), 3, seed=0)
    >>> check_cp_tensor_equal(cp_tensor, cp_tensor)
    True

    But it does not check the identity of the decompositions, only their numerical values

    >>> cp_tensor2, dataset2 = simulated_random_cp_tensor((10, 20, 30), 3, seed=0)
    >>> check_cp_tensor_equal(cp_tensor, cp_tensor2)
    True

    Normalising a ``cp_tensor`` changes its values, so then we do not have strict equality
    of the factor matrices, even though the decomposition is equivalent

    >>> from tlviz.factor_tools import normalise_cp_tensor
    >>> normalised_cp_tensor = normalise_cp_tensor(cp_tensor)
    >>> check_cp_tensor_equal(cp_tensor, normalised_cp_tensor)
    False

    Permutations will also make the numerical values of the``cp_tensor`` change

    >>> from tlviz.factor_tools import permute_cp_tensor
    >>> check_cp_tensor_equal(cp_tensor, permute_cp_tensor(cp_tensor, permutation=[1, 2, 0]))
    False

    See Also
    --------
    check_cp_tensors_equivalent : Function for checking if two CP tensors represent the same dense tensor.
    """
    validate_cp_tensor(cp_tensor1)
    validate_cp_tensor(cp_tensor2)

    rank = cp_tensor1[1][0].shape[1]
    num_modes = len(cp_tensor1[1])

    if rank != cp_tensor2[1][0].shape[1]:
        return False
    if num_modes != len(cp_tensor2[1]):
        return False

    # Check weights
    if cp_tensor1[0] is None and cp_tensor2[0] is not None:
        return False
    if cp_tensor1[0] is not None and cp_tensor2[0] is None:
        return False
    if not np.all(cp_tensor1[0] == cp_tensor2[0]):
        return False

    for mode in range(num_modes):
        if not check_factor_matrix_equal(cp_tensor1[1][mode], cp_tensor2[1][mode], ignore_labels=ignore_labels):
            return False
    return True


@_handle_tensorly_backends_dataset("factor_matrix1", None)
@_handle_tensorly_backends_dataset("factor_matrix2", None)
def check_factor_matrix_close(factor_matrix1, factor_matrix2, rtol=1e-5, atol=1e-8, ignore_labels=False):
    """Check that all entries in a factor matrix are close, if labelled, then label equality is also checked.

    This function is similar to ``numpy.allclose``, but works on both labelled and unlabelled factor
    matrices. If the factor matrices are labelled, then the DataFrame index and columns are also
    compared (unless ``ignore_labels=True``).

    Parameters
    ----------
    factor_matrix1 : numpy.ndarray or pandas.DataFrame
        Labelled or unlabelled factor matrix
    cp_tensor2 : CPTensor or tuple
        Labelled or unlabelled factor matrix
    rtol : float
        Relative tolerance (see ``numpy.allclose``)
    atol : float
        Absolute tolerance (see ``numpy.allclose``)
    ignore_labels : bool
        If True, then labels (i.e. DataFrame column names and indices) can differ.

    Returns
    -------
    bool
        Whether the decompositions are equivalent.

    Examples
    --------
    ``check_factor_matrix_close`` checks if two factor matrices are close up to round off errors.

    >>> from tlviz.data import simulated_random_cp_tensor
    >>> import numpy as np
    >>> A = np.arange(6).reshape(3, 2).astype(float)
    >>> B = A + 1e-10
    >>> check_factor_matrix_close(A, B)
    True

    If we make only one of them into a DataFrame, then the factor matrices are not close

    >>> import pandas as pd
    >>> A_labelled = pd.DataFrame(A)
    >>> check_factor_matrix_close(A_labelled, B)
    False
    >>> check_factor_matrix_close(B, A_labelled)
    False

    If we turn B into a DataFrame too, it passes again

    >>> B_labelled = pd.DataFrame(A)
    >>> check_factor_matrix_close(A_labelled, B_labelled)
    True

    The index is checked for equality, so if we change the index of ``B_labelled``, then
    the factor matrices are not close

    >>> B_labelled.index += 1
    >>> check_factor_matrix_close(A_labelled, B_labelled)
    False

    However, we can disable checking the labels by using the ``ignore_labels`` argument

    >>> check_factor_matrix_close(A_labelled, B_labelled, ignore_labels=True)
    True
    """
    if is_dataframe(factor_matrix1) != is_dataframe(factor_matrix2) and not ignore_labels:
        return False
    if ignore_labels and is_dataframe(factor_matrix1):
        factor_matrix1 = factor_matrix1.values
    if ignore_labels and is_dataframe(factor_matrix2):
        factor_matrix2 = factor_matrix2.values

    if is_dataframe(factor_matrix1):
        try:
            pd.testing.assert_frame_equal(factor_matrix1, factor_matrix2, rtol=rtol, atol=atol)
        except AssertionError:
            return False
        else:
            return True

    return np.allclose(factor_matrix1, factor_matrix2, rtol=rtol, atol=atol)


@_handle_tensorly_backends_cp("cp_tensor1", None)
@_handle_tensorly_backends_cp("cp_tensor2", None)
@_handle_none_weights_cp_tensor("cp_tensor1")
@_handle_none_weights_cp_tensor("cp_tensor2")
def check_cp_tensors_equivalent(cp_tensor1, cp_tensor2, rtol=1e-5, atol=1e-8, ignore_labels=False):
    """Check if the decompositions are equivalent

    This will check if the factor matrices and weights are equivalent. That is
    if they represent the same tensor. This differs from checking equality in
    the sense that if ``cp_tensor2`` contain the same factors as ``cp_tensor1``,
    but permuted, or with the weights distributed differently between the
    modes, then they are not equal, but equivalent.

    Parameters
    ----------
    cp_tensor1 : CPTensor or tuple
        TensorLy-style CPTensor object or tuple with weights as first
        argument and a tuple of components as second argument
    cp_tensor2 : CPTensor or tuple
        TensorLy-style CPTensor object or tuple with weights as first
        argument and a tuple of components as second argument
    rtol : float
        Relative tolerance (see ``numpy.allclose``)
    atol : float
        Absolute tolerance (see ``numpy.allclose``)
    ignore_labels : bool
        If True, then labels (i.e. DataFrame column names and indices) can differ.

    Returns
    -------
    bool
        Whether the decompositions are equivalent.

    Examples
    --------
    ``check_cp_tensors_equivalent`` checks if two CP tensors represent the same dense tensor

    >>> from tlviz.data import simulated_random_cp_tensor
    >>> from tlviz.factor_tools import check_cp_tensors_equivalent
    >>> cp_tensor, dataset = simulated_random_cp_tensor((10, 20, 30), 3, seed=0)
    >>> cp_tensor2, dataset2 = simulated_random_cp_tensor((10, 20, 30), 3, seed=0)
    >>> check_cp_tensors_equivalent(cp_tensor, cp_tensor2)
    True

    Normalising a ``cp_tensor`` changes its values, but not which dense tensor it represents

    >>> from tlviz.factor_tools import normalise_cp_tensor
    >>> normalised_cp_tensor = normalise_cp_tensor(cp_tensor)
    >>> check_cp_tensors_equivalent(cp_tensor, normalised_cp_tensor)
    True

    Permutations will also make the numerical values of the``cp_tensor`` change but not the
    dense tensor it represents

    >>> from tlviz.factor_tools import permute_cp_tensor
    >>> check_cp_tensors_equivalent(cp_tensor, permute_cp_tensor(cp_tensor, permutation=[1, 2, 0]))
    True

    See Also
    --------
    check_cp_tensor_equivalent : Function for checking if two CP tensors have the same
        numerical value (have equal weights and factor matrices)
    """
    validate_cp_tensor(cp_tensor1)
    validate_cp_tensor(cp_tensor2)

    rank = cp_tensor1[1][0].shape[1]
    num_modes = len(cp_tensor1[1])

    if rank != cp_tensor2[1][0].shape[1]:
        return False
    if num_modes != len(cp_tensor2[1]):
        return False

    for mode in range(num_modes):
        if not cp_tensor1[1][mode].shape == cp_tensor2[1][mode].shape:
            return False

    cp_tensor2 = permute_cp_tensor(cp_tensor2, reference_cp_tensor=cp_tensor1)

    cp_tensor1 = normalise_cp_tensor(cp_tensor1)
    cp_tensor2 = normalise_cp_tensor(cp_tensor2)

    if not np.allclose(cp_tensor1[0], cp_tensor2[0], rtol=rtol, atol=atol):
        return False
    for mode in range(num_modes):
        if not check_factor_matrix_close(
            cp_tensor1[1][mode], cp_tensor2[1][mode], rtol=rtol, atol=atol, ignore_labels=ignore_labels
        ):
            return False

    return True


@_handle_tensorly_backends_cp("cp_tensor", None)
@_handle_tensorly_backends_dataset("dataset", None)
@_handle_labelled_cp("cp_tensor", None)
@_handle_labelled_dataset("dataset", None, optional=True)
@_handle_none_weights_cp_tensor("cp_tensor")
def percentage_variation(cp_tensor, dataset=None, method="model"):
    r"""Compute the percentage of variation captured by each component.

    The (possible) non-orthogonality of CP factor matrices makes it less straightforward
    to estimate the amount of variation captured by each component, compared to a model with
    orthogonal factors. To estimate the amount of variation captured by a single component,
    we therefore use the following formula:

    .. math::

        \text{fit}_i = \frac{\text{SS}_i}{SS_\mathbf{\mathcal{X}}}

    where :math:`\text{SS}_i` is the squared norm of the tensor constructed using only the
    i-th component, and :math:`SS_\mathbf{\mathcal{X}}` is the squared norm of the data
    tensor :cite:p:`plstoolbox.varcap`. If ``method="data"``, then :math:`SS_\mathbf{\mathcal{X}}`
    is the squared norm of the tensor constructed from the CP tensor using all factor matrices.

    Parameters
    ----------
    cp_tensor : CPTensor or tuple
        TensorLy-style CPTensor object or tuple with weights as first
        argument and a tuple of components as second argument
    dataset : np.ndarray
        Data tensor that the cp_tensor is fitted against
    method : {"data", "model", "both"} (default="model")
        Which method to use for computing the fit.

    Returns
    -------
    fit : float or tuple
        The fit (depending on the method). If ``method="both"``, then a tuple is returned
        where the first element is the fit computed against the data tensor and the second
        element is the fit computed against the model.

    Examples
    --------
    There are two ways of computing the percentage variation. One method is to divide by the variation
    in the data, giving us the percentage variation of the data captured by each component. This
    approach will not necessarily sum to 100 since

     1. the model will not explain all the variation.
     2. the components are likely not orthogonal

    Alternatively, we can divide by the variation in the model, which will give us the contribution
    of each component to the model. However, this may also not sum to 100 since the components may
    not be orthogonal.

    >>> from tlviz.data import simulated_random_cp_tensor
    >>> from tlviz.factor_tools import percentage_variation
    >>> cp_tensor, X = simulated_random_cp_tensor((30, 10, 10), 5, noise_level=0.3, seed=0)
    >>> print(percentage_variation(cp_tensor).astype(int))
    [11  2  0  0 39]
    >>> print(percentage_variation(cp_tensor, X, method="data").astype(int))
    [11  2  0  0 37]

    We see that the variation captured for each component sums to 50 when we compare with the
    data and 52 when we compare with the model. These low numbers are because the components
    are not orthogonal, which means that the magnitude of the data is not equal to the sum
    of the magnitudes of each component. We can also compute the percentage variation with
    the model and the data simultaneously:

    >>> percent_var_data, percent_var_model = percentage_variation(cp_tensor, X, method="both")
    >>> print(percent_var_data.astype(int))
    [11  2  0  0 37]
    >>> print(percent_var_model.astype(int))
    [11  2  0  0 39]

    If noise level is 0, both methods should give the same variantion percentages:

    >>> cp_tensor, X = simulated_random_cp_tensor((30, 10, 10), 5, noise_level=0.0, seed=1)
    >>> percent_var_data, percent_var_model = percentage_variation(cp_tensor, X, method="both")
    >>> print(percent_var_data.astype(int))
    [ 3 11  0 34  1]
    >>> print(f"Sum of variation: {percent_var_data.sum():.0f}")
    Sum of variation: 51
    >>> print(percent_var_model.astype(int))
    [ 3 11  0 34  1]
    >>> print(f"Sum of variation: {percent_var_model.sum():.0f}")
    Sum of variation: 51
    """
    weights, factor_matrices = cp_tensor
    ssc = weights ** 2

    if dataset is not None and method == "model":
        warn(
            'Dataset provided but method="model", so it is not used. To compute the variation'
            + ' captured in the data, use method="data" or method="both".'
        )

    for factor_matrix in factor_matrices:
        ssc = ssc * np.sum(np.abs(factor_matrix) ** 2, axis=0)

    if method == "data":
        if dataset is None:
            raise TypeError("The dataset must be provided if method='data'")
        return 100 * ssc / np.sum(dataset ** 2)
    elif method == "model":
        return 100 * ssc / (cp_norm(cp_tensor) ** 2)
    elif method == "both":
        return 100 * ssc / np.sum(dataset ** 2), 100 * ssc / (cp_norm(cp_tensor) ** 2)
    else:
        raise ValueError("Method must be either 'data', 'model' or 'both")



# -*- coding: utf-8 -*-

__author__ = "Marie Roald & Yngve Mardal Moe"

import inspect
from functools import wraps

import numpy as np
import xarray as xr

from ._module_utils import _handle_none_weights_cp_tensor, validate_cp_tensor
from ._tl_utils import _handle_tensorly_backends_cp, _handle_tensorly_backends_dataset, _SINGLETON
from ._xarray_wrapper import (
    _handle_labelled_cp,
    _handle_labelled_dataset,
    is_labelled_cp,
    is_labelled_dataset,
    is_labelled_tucker,
)

__all__ = [
    "extract_singleton",
    "unfold_tensor",
    "cp_to_tensor",
    "tucker_to_tensor",
    "normalise",
    "is_labelled_cp",
    "is_labelled_tucker",
    "is_labelled_dataset",
]


def _alias_mode_axis():
    def decorator(func):
        func_sig = inspect.signature(func)
        if "axis" not in func_sig.parameters or "mode" not in func_sig.parameters:
            raise TypeError(f"Function {func} needs both ``mode`` and ``axis`` as possible arguments.")
        mode_default_value = func_sig.parameters["mode"].default
        if mode_default_value == inspect._empty:
            mode_default_value = None

        @wraps(func)
        def new_func(*args, **kwargs):
            bound_arguments = func_sig.bind_partial(*args, **kwargs)

            mode = bound_arguments.arguments.get("mode", mode_default_value)
            axis = bound_arguments.arguments.get("axis", None)
            if mode is None and axis is None:
                raise TypeError(
                    f"Function {func} needs either ``mode`` or ``axis`` to be set to a value different than None."
                )
            elif mode != mode_default_value and axis is not None:
                raise TypeError("Either ``mode`` or ``axis`` can be specified, not both.")
            elif axis is not None:
                bound_arguments.arguments["mode"] = axis
            return func(**bound_arguments.arguments)

        return new_func

    return decorator


def extract_singleton(x):
    """Extracts a singleton from an array.

    This is useful whenever xarray or Pandas is used, since many NumPy functions that
    return a number may return a singleton array instead.

    Parameters
    ----------
    x : float, numpy.ndarray, xarray.DataArray or pandas.DataFrame
        Singleton array to extract value from.

    Returns
    -------
    float
        Singleton value extracted from ``x``.
    """
    return np.asarray(x).reshape(-1).item()


@_handle_tensorly_backends_cp("cp_tensor", None)
@_handle_none_weights_cp_tensor("cp_tensor")
@_handle_labelled_cp("cp_tensor", None)
def cp_norm(cp_tensor):
    r"""Efficiently compute the Frobenius norm of a possibly labelled CP tensor.

    The norm for an :math:`R`-component :math:`I \times J \times K` CP tensor is
    computed in :math:`O((I + J + K)R^2)` flops instead of :math:`O(IJKR)` flops.

    Parameters
    ----------
    cp_tensor : CPTensor or tuple
        TensorLy-style CPTensor object or tuple with weights as first
        argument and a tuple of components as second argument.

    Returns
    -------
    float : The norm of the CP tensor

    Examples
    --------
    >>> import numpy as np
    >>> from tlviz.data import simulated_random_cp_tensor
    >>> from tlviz.utils import cp_norm, cp_to_tensor
    >>> cp_tensor = simulated_random_cp_tensor((30, 10, 10), rank=5, seed=0)[0]
    >>> norm_fast = cp_norm(cp_tensor)
    >>> norm_slow = np.linalg.norm(cp_to_tensor(cp_tensor))
    >>> print(f"Difference in norm: {abs(norm_fast - norm_slow):.2f}")
    Difference in norm: 0.00
    """
    weights, factor_matrices = cp_tensor
    model_norm = weights[:, np.newaxis] * weights[np.newaxis, :]

    for fm in factor_matrices:
        model_norm *= np.conjugate(fm.T) @ fm

    return np.sqrt(model_norm.sum())


@_handle_tensorly_backends_dataset("tensor", _SINGLETON)
@_handle_labelled_dataset("tensor", None)
@_alias_mode_axis()
def unfold_tensor(tensor, mode, axis=None):
    """Unfolds (matricises) a potentially labelled data tensor into a numpy array along given mode.

    Arguments
    ---------
    tensor : np.ndarray or xarray.DataArray
        Dataset to unfold
    mode : int
        Which mode (axis) to unfold the dataset along.
    axis : int (optional)
        Which mode (axis) to unfold the dataset along. If set, then the mode-argument is unused.

    Returns
    -------
    np.ndarray
        The unfolded dataset as a numpy array.
    """
    # TODO: return xarray or dataframe if tensor is labelled
    dataset = np.asarray(tensor)
    return np.moveaxis(dataset, mode, 0).reshape(dataset.shape[mode], -1)


@_handle_tensorly_backends_cp("cp_tensor", None)
def cp_to_tensor(cp_tensor):
    """Convert a CP tensor to a dense array.

    This function is equivalent to ``cp_to_tensor`` in TensorLy, but supports dataframes.

    If the factor matrices are DataFrames, then the tensor will be returned as a labelled
    xarray. Otherwise, it will be returned as a numpy array.

    Parameters
    ----------
    cp_tensor : CPTensor or tuple
        TensorLy-style CPTensor object or tuple with weights as first
        argument and a tuple of components as second argument.

    Returns
    -------
    xarray or np.ndarray
        Dense tensor represented by the decomposition.

    Examples
    --------
    >>> from tlviz.data import simulated_random_cp_tensor
    >>> from tlviz.utils import cp_to_tensor
    >>> unlabelled_cp_tensor = simulated_random_cp_tensor((10, 20, 30), 5, labelled=False, seed=0)[0]
    >>> unlabelled_X = cp_to_tensor(unlabelled_cp_tensor)
    >>> type(unlabelled_X)
    <class 'numpy.ndarray'>
    >>> labelled_cp_tensor = simulated_random_cp_tensor((10, 20, 30), 5, labelled=True, seed=0)[0]
    >>> labelled_X = cp_to_tensor(labelled_cp_tensor)
    >>> type(labelled_X)
    <class 'xarray.core.dataarray.DataArray'>
    >>> np.array_equal(unlabelled_X, labelled_X.data)
    True
    """
    cp_tensor = validate_cp_tensor(cp_tensor)

    if cp_tensor[0] is None:
        weights = np.ones(cp_tensor[1][0].shape[1])
    else:
        weights = cp_tensor[0].reshape(-1)

    einsum_input = "R"
    einsum_output = ""
    for mode in range(len(cp_tensor[1])):
        idx = chr(ord("a") + mode)

        # We cannot use einsum with letters outside the alphabet
        if ord(idx) > ord("z"):
            max_modes = ord("a") - ord("z") - 1
            raise ValueError(f"Cannot have more than {max_modes} modes. Current components have {len(cp_tensor[1])}.")

        einsum_input += f", {idx}R"
        einsum_output += idx

    tensor = np.einsum(f"{einsum_input} -> {einsum_output}", weights, *cp_tensor[1])

    if not is_labelled_cp(cp_tensor):
        return tensor

    # Convert to labelled xarray DataArray:
    coords_dict = {}
    dims = []
    for mode, fm in enumerate(cp_tensor[1]):
        mode_name = f"Mode {mode}"
        if fm.index.name is not None:
            mode_name = fm.index.name

        coords_dict[mode_name] = fm.index.values
        dims.append(mode_name)

    return xr.DataArray(tensor, dims=dims, coords=coords_dict)


@_handle_tensorly_backends_cp("tucker_tensor", None)
def tucker_to_tensor(tucker_tensor):
    """Convert a Tucker tensor to a dense array.

    This function is equivalent to ``tucker_to_tensor`` in TensorLy, but supports dataframes.

    If the factor matrices are DataFrames, then the tensor will be returned as a labelled
    xarray. Otherwise, it will be returned as a numpy array.

    Parameters
    ----------
    tucker : CPTensor or tuple
        TensorLy-style TuckerTensor object or tuple with weights as first
        argument and a tuple of components as second argument.

    Returns
    -------
    xarray or np.ndarray
        Dense tensor represented by the decomposition.
    """
    einsum_core = ""
    einsum_input = ""
    einsum_output = ""
    if len(tucker_tensor[1]) > 16:
        raise ValueError("NumPy's einsum function doesn't support forming dense Tucker arrays with more than 16 modes.")

    for mode in range(len(tucker_tensor[1])):
        idx = chr(ord("a") + mode)
        rank_idx = chr(ord("A") + mode)

        einsum_core += rank_idx
        einsum_input += f", {idx}{rank_idx}"
        einsum_output += idx

    tensor = np.einsum(f"{einsum_core}{einsum_input} -> {einsum_output}", tucker_tensor[0], *tucker_tensor[1],)
    if not is_labelled_tucker(tucker_tensor):
        return tensor

    # Convert to labelled xarray DataArray:
    coords_dict = {}
    dims = []
    for mode, fm in enumerate(tucker_tensor[1]):
        mode_name = f"Mode {mode}"
        if fm.index.name is not None:
            mode_name = fm.index.name

        coords_dict[mode_name] = fm.index.values
        dims.append(mode_name)

    return xr.DataArray(tensor, dims=dims, coords=coords_dict)


@_alias_mode_axis()
@_handle_tensorly_backends_dataset("x", _SINGLETON)
@_handle_labelled_dataset("x", _SINGLETON)
def normalise(x, mode=0, axis=None):
    """Normalise a matrix (or tensor) so all columns (or fibers) have unit norm.

    Parameters
    ----------
    x : np.ndarray
        Matrix (or vector/tensor) to normalise.
    mode : int
        Axis along which to normalise, if 0, then all columns will have unit norm
        and if 1 then all rows will have unit norm. When normalising a tensor, then
        the axis represents the axis whose fibers should have unit norm.
    axis : int
        Alias for mode. If this is provided, then no value for mode can be provided.

    Returns
    -------
    np.ndarray
        Normalised matrix

    Examples
    --------
    >>> random_matrix = np.random.random_sample((3, 4))
    >>> matrix_normalized_cols = normalise(random_matrix, axis=0)
    >>> print(np.linalg.norm(matrix_normalized_cols, axis=0))
    [1. 1. 1. 1.]

    >>> random_matrix = np.random.random_sample((3, 4))
    >>> matrix_normalized_rows = normalise(random_matrix, axis=1)
    >>> print(np.linalg.norm(matrix_normalized_rows, axis=1))
    [1. 1. 1.]
    """
    norms = np.linalg.norm(x, axis=mode, keepdims=True)
    norms[norms == 0] = 1
    return x / norms



# -*- coding: utf-8 -*-

__author__ = "Marie Roald & Yngve Mardal Moe"

"""
This module contains functions used to evaluate a single tensor factorisation model
by comparing it to a data tensor.
"""
import numpy as np
import scipy.linalg as sla

from ._tl_utils import _handle_tensorly_backends_cp, _handle_tensorly_backends_dataset, to_numpy
from ._xarray_wrapper import _handle_labelled_cp, _handle_labelled_dataset
from .utils import _alias_mode_axis, cp_to_tensor

__all__ = [
    "estimate_core_tensor",
    "core_consistency",
    "sse",
    "relative_sse",
    "fit",
    "predictive_power",
]


@_handle_tensorly_backends_dataset("dataset", None)
@_handle_labelled_dataset("dataset", None)
def estimate_core_tensor(factors, dataset):
    """Efficient estimation of the Tucker core from a factor matrices and a data tensor.

    Parameters
    ----------
    factors : tuple
        Tuple of factor matrices used to estimate the core tensor from
    dataset : np.ndarray
        The data tensor that the core tensor is estimated from

    Notes
    -----
    In the original paper, :cite:t:`papalexakis2015fast` present an algorithm
    for 3-way tensors. However, it is straightforward to generalise it to N-way tensors
    by using the inverse tensor product formula in :cite:p:`buis1996efficient`.
    """
    factors = [to_numpy(factor, cast_labelled=True) for factor in factors]

    svds = [sla.svd(factor, full_matrices=False) for factor in factors]
    for U, s, Vh in svds[::-1]:
        dataset = np.tensordot(U.T, dataset, (1, dataset.ndim - 1))
    for U, s, Vh in svds[::-1]:
        s_pinv = s.copy()
        mask = s_pinv != 0
        s_pinv[mask] = 1 / s_pinv[mask]
        dataset = np.tensordot(np.diag(s_pinv), dataset, (1, dataset.ndim - 1))
    for U, s, Vh in svds[::-1]:
        dataset = np.tensordot(Vh.T, dataset, (1, dataset.ndim - 1))
    return np.ascontiguousarray(dataset)


@_handle_tensorly_backends_dataset("dataset", None)
@_handle_tensorly_backends_cp("cp_tensor", None)
@_handle_labelled_dataset("dataset", None)
@_handle_labelled_cp("cp_tensor", None)
def core_consistency(cp_tensor, dataset, normalised=False):
    r"""Computes the core consistency :cite:p:`bro2003new`

    A CP model can be interpreted as a restricted Tucker model, where the
    core tensor is constrained to be superdiagonal. For a third order tensor,
    this means that the core tensor, :math:`\mathcal{G}`, satisfy :math:`g_{ijk}\neq0`
    only if :math:`i = j = k`. To compute the core consistency of a CP decomposition,
    we use this property, and calculate the optimal Tucker core tensor given
    the factor matrices of the CP model.

    The key observation is that if the data tensor follows the assumptions
    of the CP model, then the optimal core tensor should be similar to that
    of the CP model, i. e. superdiagonal. However, if the data can be better
    described by allowing for interactions between the components across modes,
    then the core tensor will have non-zero off-diagonal. The core consistency
    quantifies this measure and is defined as:

    .. math::

        \text{CC} = 100 - 100 \frac{\| \mathcal{G} - \mathcal{I} \|_F^2}{N}

    where :math:`\mathcal{G}` is the estimated core tensor, :math:`\mathcal{I}`
    is a superdiagonal tensor only ones on the superdiagonal and :math:`N`
    is a normalising factor, either equal to the number of components or the
    squared frobenius norm of the estimated core tensor. A core consistency
    score close to 100 indicates that the CP model is likely valid. If the
    core consistency is low, however, then the model either has components
    that describe noise or the data does not follow the model's assumptions.
    So the core consistency can help determine if the chosen number of
    components is suitable.

    Parameters
    ----------
    cp_tensor : CPTensor or tuple
        TensorLy-style CPTensor object or tuple with weights as first
        argument and a tuple of components as second argument
    dataset : np.ndarray
        Data tensor that the cp_tensor is fitted against
    normalised : Bool (default=False)
        If True, then the squared frobenius norm of the estimated core tensor
        is used to normalise the core consistency. Otherwise, the number of
        components is used.

        If ``normalised=False``, then the core consistency formula coincides
        with :cite:p:`bro2003new`, and if ``normalised=True``, the core consistency
        formula coincides with that used in the `N-Way toolbox <http://models.life.ku.dk/nwaytoolbox>`_,
        and is unlikely to be less than 0. For core consistencies close to
        100, the formulas approximately coincide.

    Returns
    -------
    float
        The core consistency

    Examples
    --------
    We can use the core consistency diagonstic to determine the correct number of components
    for a CP model. Here, we only fit one model, but in practice, you should fit multiple models
    and select the one with the lowest SSE (to account for local minima) before computing the
    core consistency.

    >>> from tlviz.data import simulated_random_cp_tensor
    >>> from tensorly.decomposition import parafac
    >>> cp_tensor, dataset = simulated_random_cp_tensor((10,11,12), 3, seed=42)
    >>> # Fit many CP models with different number of components
    >>> for rank in range(1, 5):
    ...     decomposition = parafac(dataset, rank=rank, random_state=42)
    ...     cc = core_consistency(decomposition, dataset, normalised=True)
    ...     print(f"No. components: {rank} - core consistency: {cc:.0f}")
    No. components: 1 - core consistency: 100
    No. components: 2 - core consistency: 100
    No. components: 3 - core consistency: 81
    No. components: 4 - core consistency: 0

    .. note::
    
        This implementation uses the fast method of estimating the core tensor
        :cite:p:`papalexakis2015fast,buis1996efficient`
    """
    # Distribute weights
    weights, factors = cp_tensor
    rank = factors[0].shape[1]

    A = factors[0].copy()
    if weights is not None:
        A *= weights.reshape(1, -1)

    factors = tuple((A, *factors[1:]))

    # Estimate core and compare
    G = estimate_core_tensor(factors, dataset)
    T = np.zeros([rank] * dataset.ndim)
    np.fill_diagonal(T, 1)
    if normalised:
        denom = np.sum(G ** 2)
    else:
        denom = rank

    return 100 - 100 * np.sum((G - T) ** 2) / denom


@_handle_tensorly_backends_dataset("dataset", None)
@_handle_tensorly_backends_cp("cp_tensor", None)
@_handle_labelled_dataset("dataset", None)
@_handle_labelled_cp("cp_tensor", None)
def sse(cp_tensor, dataset):
    """Compute the sum of squared error for a given cp_tensor.

    Parameters
    ----------
    cp_tensor : CPTensor or tuple
        TensorLy-style CPTensor object or tuple with weights as first
        argument and a tuple of components as second argument
    dataset : ndarray
        Tensor approximated by ``cp_tensor``

    Returns
    -------
    float
        The sum of squared error, ``sum((X_hat - dataset)**2)``, where ``X_hat``
        is the dense tensor represented by ``cp_tensor``

    Examples
    --------
    Below, we create a random CP tensor and a random tensor and compute
    the sum of squared error for these two tensors.

    >>> import tensorly as tl
    >>> from tensorly.random import random_cp
    >>> from tlviz.model_evaluation import sse
    >>> rng = tl.check_random_state(0)
    >>> cp = random_cp((4, 5, 6), 3, random_state=rng)
    >>> X = rng.random_sample((4, 5, 6))
    >>> sse(cp, X)
    18.948918157419186
    """
    X_hat = cp_to_tensor(cp_tensor)
    return np.sum((dataset - X_hat) ** 2)


@_handle_tensorly_backends_dataset("dataset", None)
@_handle_tensorly_backends_cp("cp_tensor", None)
@_handle_labelled_dataset("dataset", None)
@_handle_labelled_cp("cp_tensor", None)
def relative_sse(cp_tensor, dataset, sum_squared_dataset=None):
    """Compute the relative sum of squared error for a given cp_tensor.

    Parameters
    ----------
    cp_tensor : CPTensor or tuple
        TensorLy-style CPTensor object or tuple with weights as first
        argument and a tuple of components as second argument
    dataset : ndarray
        Tensor approximated by ``cp_tensor``
    sum_squared_dataset: float (optional)
        If ``sum(dataset**2)`` is already computed, you can optionally provide it
        using this argument to avoid unnecessary recalculation.

    Returns
    -------
    float
        The relative sum of squared error, ``sum((X_hat - dataset)**2)/sum(dataset**2)``,
        where ``X_hat`` is the dense tensor represented by ``cp_tensor``

    Examples
    --------
    Below, we create a random CP tensor and a random tensor and compute
    the sum of squared error for these two tensors.

    >>> import tensorly as tl
    >>> from tensorly.random import random_cp
    >>> from tlviz.model_evaluation import relative_sse
    >>> rng = tl.check_random_state(0)
    >>> cp = random_cp((4, 5, 6), 3, random_state=rng)
    >>> X = rng.random_sample((4, 5, 6))
    >>> relative_sse(cp, X)
    0.4817407254961442
    """
    if sum_squared_dataset is None:
        sum_squared_x = np.sum(dataset ** 2)
    return sse(cp_tensor, dataset) / sum_squared_x


@_handle_tensorly_backends_dataset("dataset", None)
@_handle_tensorly_backends_cp("cp_tensor", None)
@_handle_labelled_dataset("dataset", None)
@_handle_labelled_cp("cp_tensor", None)
def fit(cp_tensor, dataset, sum_squared_dataset=None):
    """Compute the fit (1-relative sum squared error) for a given cp_tensor.

    Parameters
    ----------
    cp_tensor : CPTensor or tuple
        TensorLy-style CPTensor object or tuple with weights as first
        argument and a tuple of components as second argument
    dataset : ndarray
        Tensor approximated by ``cp_tensor``
    sum_squared_dataset: float (optional)
        If ``sum(dataset**2)`` is already computed, you can optionally provide it
        using this argument to avoid unnecessary recalculation.

    Returns
    -------
    float
        The relative sum of squared error, ``sum((X_hat - dataset)**2)/sum(dataset**2)``,
        where ``X_hat`` is the dense tensor represented by ``cp_tensor``

    Examples
    --------
    Below, we create a random CP tensor and a random tensor and compute
    the sum of squared error for these two tensors.

    >>> import tensorly as tl
    >>> from tensorly.random import random_cp
    >>> from tlviz.model_evaluation import fit
    >>> rng = tl.check_random_state(0)
    >>> cp = random_cp((4, 5, 6), 3, random_state=rng)
    >>> X = rng.random_sample((4, 5, 6))
    >>> fit(cp, X)
    0.5182592745038558

    We can see that it is equal to 1 - relative SSE

    >>> from tlviz.model_evaluation import relative_sse
    >>> 1 - relative_sse(cp, X)
    0.5182592745038558
    """
    return 1 - relative_sse(cp_tensor, dataset, sum_squared_dataset=sum_squared_dataset)


@_alias_mode_axis()
@_handle_tensorly_backends_cp("cp_tensor", None)
def predictive_power(cp_tensor, y, sklearn_estimator, mode=0, metric=None, axis=None):
    """Use scikit-learn estimator to evaluate the predictive power of a factor matrix.

    This is useful if you evaluate the components based on their predictive
    power with respect to some task.

    Parameters
    ----------
    factor_matrix : ndarray(ndim=2)
        Factor matrix from a tensor decomposition model
    y : ndarray(ndim=1)
        Prediction target for each row of the factor matrix in the given mode.
        ``y`` should have same length as the first dimension of this factor
        matrix (i.e. the length of the tensor along the given mode).
    sklearn_estimator : scikit learn estimator
        Scikit learn estimator. Must have the ``fit`` and ``predict`` methods,
        and if ``metric`` is ``None``, then it should also have the ``score``
        method. See https://scikit-learn.org/stable/developers/develop.html.
    mode : int
        Which mode to perform the scoring along
    metric : Callable
        Callable (typically function) with the signature ``metric(y_true, y_pred)``,
        where ``y_true=labels`` and ``y_pred`` is the predicted values
        obtained from ``sklearn_estimator``. See
        https://scikit-learn.org/stable/developers/develop.html#specific-models.
    axis : int (optional)
        Alias for mode, if set, then mode cannot be set.

    Returns
    -------
    float
        Score based on the estimator's performance.

    Examples
    --------
    ``predictive_power`` can be useful to evaluate the predictive power of a CP decomposition.
    To illustrate this, we start by creating a simulated CP tensor and a variable we want to
    predict that is linearly related to one of the factor matrices.

    >>> from tlviz.data import simulated_random_cp_tensor
    >>> import numpy as np
    >>> rng = np.random.default_rng(0)
    >>> cp_tensor, X = simulated_random_cp_tensor((30, 10, 10), 5, noise_level=0.3, seed=rng)
    >>> weights, (A, B, C) = cp_tensor
    >>> regression_coefficients = rng.standard_normal((5, 1))
    >>> Y = A @ regression_coefficients

    Next, we fit a PARAFAC model to this data

    >>> from tensorly.decomposition import parafac
    >>> est_cp_tensor = parafac(X, 5)

    Finally, we see how well the estimated decomposition can describe our target variable, ``Y``.
    This will use the :math:`R^2`-coefficient for scoring, as that is the default scoring method
    for linear models.

    >>> from sklearn.linear_model import LinearRegression
    >>> from tlviz.model_evaluation import predictive_power
    >>> linear_regression = LinearRegression()
    >>> r_squared = predictive_power(cp_tensor, Y, linear_regression)
    >>> print(f"The R^2 coefficient is {r_squared:.2f}")
    The R^2 coefficient is 1.00

    We can also specify our own scoring function

    >>> from sklearn.metrics import max_error
    >>> highest_error = predictive_power(cp_tensor, Y, linear_regression, metric=max_error)
    >>> print(f"The maximum error is {highest_error:.2f}")
    The maximum error is 0.00
    """
    factor_matrix = cp_tensor[1][mode]
    sklearn_estimator.fit(factor_matrix, y)
    if metric is None:
        return sklearn_estimator.score(factor_matrix, y)
    return metric(y, sklearn_estimator.predict(factor_matrix))



from functools import wraps
from inspect import signature
from textwrap import dedent

try:
    import tensorly as tl
except ImportError:
    HAS_TENSORLY = False
else:
    HAS_TENSORLY = True

from ._module_utils import _SINGLETON, _check_is_argument
from ._xarray_wrapper import (
    is_labelled_cp,
    is_labelled_dataset,
)


def _all_not(iterable):
    return all(not x for x in iterable)


def is_tensorly_cp(cp_tensor, none_ok=False):
    if cp_tensor is None and none_ok:
        return False
    elif cp_tensor is None:
        raise TypeError("cp_tensor is None, but none_ok=False")

    if not HAS_TENSORLY:
        return False

    weights, factors = cp_tensor
    tl_weights = tl.is_tensor(weights) or weights is None
    tl_factors = [tl.is_tensor(factor) for factor in factors]

    if tl_weights and all(tl_factors):
        return True
    elif not tl_weights and _all_not(tl_factors):
        return False
    elif is_labelled_cp(cp_tensor):
        return False
    else:
        raise TypeError(
            dedent(
                f"""\
                CP tensor has mixed tensorly and non-tensorly types.
                The weights have type: {type(weights)}
                The factors have types: {[type(f) for f in factors]}
                """.rstrip()
            )
        )


def to_numpy_cp(cp_tensor, cast_labelled_cp=True):
    if is_labelled_cp(cp_tensor) and not cast_labelled_cp:
        return cp_tensor
    elif is_labelled_cp(cp_tensor):
        weights, factors = cp_tensor
        return weights, [factor.values for factor in factors]

    if not is_tensorly_cp(cp_tensor):
        return cp_tensor

    weights, factors = cp_tensor
    if weights is not None:
        weights = tl.to_numpy(weights)
    factors = [tl.to_numpy(factor_matrix) for factor_matrix in factors]

    return weights, factors


def to_numpy(x, cast_labelled=True):
    if is_labelled_dataset(x) and not cast_labelled:
        return x
    elif is_labelled_dataset(x):
        return x.values

    if not HAS_TENSORLY:
        return x
    return tl.to_numpy(x)


def _handle_tensorly_backends_cp(cp_tensor_name, output_cp_tensor_index, optional=False):
    # TOTEST: _handle_tensorly_backends_cp
    def decorator(func):
        _check_is_argument(func, cp_tensor_name)

        @wraps(func)
        def func2(*args, **kwargs):
            if not HAS_TENSORLY:
                return func(*args, **kwargs)

            bound_arguments = signature(func).bind(*args, **kwargs)

            cp_tensor = bound_arguments.arguments.get(cp_tensor_name, None)
            if cp_tensor is None and optional:
                np_cp_tensor = None
            elif cp_tensor is None:
                raise ValueError("cp_tensor is None, but it is not optional.")
            else:
                np_cp_tensor = to_numpy_cp(cp_tensor, cast_labelled_cp=False)

            bound_arguments.arguments[cp_tensor_name] = np_cp_tensor
            out = func(*bound_arguments.args, **bound_arguments.kwargs)

            if output_cp_tensor_index is _SINGLETON and is_tensorly_cp(cp_tensor, none_ok=optional):
                weights, factors = out
                if weights is not None:
                    weights = tl.tensor(weights)
                return tl.cp_tensor.CPTensor((weights, [tl.tensor(factor) for factor in factors]))
            elif output_cp_tensor_index is not None and is_tensorly_cp(cp_tensor, none_ok=optional):
                weights, factors = out[output_cp_tensor_index]
                if weights is not None:
                    weights = tl.tensor(weights)
                out_cp_tensor = tl.cp_tensor.CPTensor((weights, [tl.tensor(factor) for factor in factors]))

                return (
                    *out[:output_cp_tensor_index],
                    out_cp_tensor,
                    *out[output_cp_tensor_index + 1 :],
                )
            return out

        return func2

    return decorator


def _handle_tensorly_backends_dataset(dataset_name, output_dataset_index):
    def decorator(func):
        _check_is_argument(func, dataset_name)

        @wraps(func)
        def func2(*args, **kwargs):
            if not HAS_TENSORLY:
                return func(*args, **kwargs)

            bound_arguments = signature(func).bind(*args, **kwargs)

            dataset = bound_arguments.arguments.get(dataset_name, None)
            if tl.is_tensor(dataset):
                is_tensorly = True
                np_dataset = tl.to_numpy(dataset)
            else:
                is_tensorly = False
                np_dataset = dataset

            bound_arguments.arguments[dataset_name] = np_dataset
            out = func(*bound_arguments.args, **bound_arguments.kwargs)

            if output_dataset_index is _SINGLETON and is_tensorly:
                return tl.tensor(out)
            elif output_dataset_index is not None and is_tensorly:
                out_dataset = tl.tensor(out[output_dataset_index])

                return (
                    *out[:output_dataset_index],
                    out_dataset,
                    *out[output_dataset_index + 1 :],
                )
            return out

        return func2

    return decorator



# -*- coding: utf-8 -*-

__author__ = "Marie Roald & Yngve Mardal Moe"

import matplotlib.pyplot as plt
import pytest


def pytest_configure(config):
    import matplotlib

    matplotlib.use("pdf")


@pytest.fixture(autouse=True)
def close_matplotlib_figures():
    yield
    plt.close("all")



# -*- coding: utf-8 -*-

__author__ = "Marie Roald & Yngve Mardal Moe"

"""
Utilities for comparing multiple decompositions against each other.
"""
import numpy as np

from . import model_evaluation
from .factor_tools import factor_match_score
from .utils import extract_singleton

__all__ = [
    "similarity_evaluation",
    "get_model_with_lowest_error",
    "sort_models_by_error",
]


def similarity_evaluation(cp_tensor, comparison_cp_tensors, similarity_metric=None, **kwargs):
    r"""Compute similarities between ``cp_tensor`` and all ``comparison_cp_tensors``.

    Parameters
    ----------
    cp_tensor : CPTensor or tuple
        TensorLy-style CPTensor object or tuple with weights as first
        argument and a tuple of components as second argument
    comparison_cp_tensors : List[CPTensor or tuple]
        List of TensorLy-style CPTensors to compare with
    similarity_metric : Callable[CPTensor, CPTensor, \*\*kwargs] -> float
        Function that takes two CPTensors as input and returns their similarity
    **kwargs
        Extra keyword-arguments passed to ``similarity_metric``.

    Returns
    -------
    similarity : float

    Examples
    --------
    In this example, we will fit several PARAFAC models to a simulated dataset and use ``similarity_evaluation`` to
    compute the similarities between the different fitted models and the model that obtained the lowest error.

    We start by importing the relevant functionality

    >>> from tlviz.multimodel_evaluation import sort_models_by_error, similarity_evaluation
    >>> from tlviz.data import simulated_random_cp_tensor
    >>> from tensorly.decomposition import parafac

    Then, we create a random simulated dataset and fit five parafac models to it.

    >>> cp_tensor, dataset = simulated_random_cp_tensor((10, 20, 30), 3, seed=0)
    >>> model_candidates = [
    ...     parafac(dataset, 3, init="random", random_state=i)
    ...     for i in range(5)
    ... ]

    Finally, we sort the models by their errors and compute the similarity between each model
    and the model that obtained the lowest error.

    >>> sorted_model_candidates, errors = sort_models_by_error(model_candidates, dataset)
    >>> similarities = similarity_evaluation(sorted_model_candidates[0], sorted_model_candidates[1:])
    >>> for i, s in enumerate(similarities):
    ...     print(f"Similarity between the model with the lowest loss and the model with the {i+2}. lowest loss: {s:.2}")
    Similarity between the model with the lowest loss and the model with the 2. lowest loss: 0.99
    Similarity between the model with the lowest loss and the model with the 3. lowest loss: 0.98
    Similarity between the model with the lowest loss and the model with the 4. lowest loss: 0.68
    Similarity between the model with the lowest loss and the model with the 5. lowest loss: 0.42

    We see that the three models with the lowest error were very similar, which indicates that the model is stable.
    """
    if similarity_metric is None:
        similarity_metric = factor_match_score

    return [
        similarity_metric(cp_tensor, comparison_cp_tensor, **kwargs) for comparison_cp_tensor in comparison_cp_tensors
    ]


def get_model_with_lowest_error(cp_tensors, dataset, error_function=None, return_index=False, return_errors=False):
    """Compute reconstruction error for all cp_tensors and return model with lowest error.

    This is useful to select the best initialisation if several random
    initialisations are used to fit the model. By default, the relative SSE
    is used, but another error function can be used too.

    Parameters
    ----------
    cp_tensors : list of CPTensors
        List of all CP tensors to compare
    dataset : ndarray
        Dataset modelled by the CP tensors
    error_function : Callable (optional)
        Callable with the signature ``error_function(cp_tensor, dataset)``,
        that should return a measure of the modelling error (e.g. SSE). Default
        is relative SSE.
    return_index : bool (optional, default=False)
        If True, then the index of the CP tensor with the lowest error is returned
    return_errors : bool (optional, defult=False)
        if True, then a list of errors for each CP tensor is returned.

    Returns
    -------
    CPTensor
        The CP tensor with the lowest error
    int
        The index of the selected CP tensor in ``cp_tensors``. Only returned
        if ``return_index=True``.
    list
        List of the error values for all CP tensors in ``cp_tensor`` (in the same
        order as ``cp_tensors``). only returned if ``return_errors=True``

    Examples
    --------
    Here, we illustrate how ``get_model_with_lowest_error`` can be used to get the selected model from a collection
    of model candidates, and how we can also get the errors for all model candidates and the index of the selected
    initialisation.

    We start by importing the relevant functionality

    >>> from tlviz.multimodel_evaluation import sort_models_by_error, get_model_with_lowest_error
    >>> from tlviz.model_evaluation import relative_sse
    >>> from tlviz.data import simulated_random_cp_tensor
    >>> from tlviz.factor_tools import check_cp_tensor_equal
    >>> from tensorly.decomposition import parafac

    Then, we create a simulated dataset and fit five model candidates using different random initialisations.

    >>> cp_tensor, dataset = simulated_random_cp_tensor((10, 20, 30), 3, noise_level=0.3, seed=0)
    >>> model_candidates = [
    ...     parafac(dataset, 3, init="random", random_state=i)
    ...     for i in range(5)
    ... ]

    Once we have the model candidates, we use ``get_model_with_lowest_error``. By default this function will only
    return the selected model, but in this case, we ask it to return the index of the selected model and the errors
    of all model candidates.

    >>> model, index, errors = get_model_with_lowest_error(model_candidates, dataset, return_index=True, return_errors=True)
    >>> print(f"Model {index} has lowest error")
    Model 3 has lowest error

    We can check that the selected model is the model with the init we got

    >>> check_cp_tensor_equal(model, model_candidates[index])
    True

    And that it is the model that has the lowest error

    >>> errors[index] == min(errors)
    True

    And finally that this error is equal to the relative SSE

    >>> errors[index] == relative_sse(model, dataset)
    True
    """
    if error_function is None:
        error_function = model_evaluation.relative_sse

    selected_cp_tensor = None
    selected_index = None
    lowest_sse = np.inf
    all_sse = []
    for i, cp_tensor in enumerate(cp_tensors):
        sse = error_function(cp_tensor, dataset)
        all_sse.append(sse)
        if sse < lowest_sse:
            selected_cp_tensor = cp_tensor
            lowest_sse = sse
            selected_index = i

    returns = [selected_cp_tensor]
    if return_index:
        returns.append(selected_index)
    if return_errors:
        returns.append(all_sse)
    returns = tuple(returns)
    if len(returns) == 1:
        return returns[0]
    else:
        return returns


def sort_models_by_error(cp_tensors, dataset, error_function=None):
    """Sort the ``cp_tensors`` by their error so the model with the lowest error is first.

    Parameters
    ----------
    cp_tensors : list of CPTensors
        List of all CP tensors
    dataset : ndarray
        Dataset modelled by the CP tensors
    error_function : Callable (optional)
        Callable with the signature ``error_function(cp_tensor, dataset)``,
        that should return a measure of the modelling error (e.g. SSE).

    Returns
    -------
    list of CPTensors
        List of all CP tensors sorted so the CP tensor with the lowest error
        is first and highest error is last.
    list of floats
        List of error computed for each CP tensor (in sorted order)

    Examples
    --------
    Here, we see how ``sort_models_by_error`` can be useful to get a collection of model candidates in a logical order.

    We start by importing the relevant functionality.

    >>> from tlviz.multimodel_evaluation import sort_models_by_error, get_model_with_lowest_error
    >>> from tlviz.data import simulated_random_cp_tensor
    >>> from tensorly.decomposition import parafac

    Then, we simulate a random dataset and fit five model candidates to it.

    >>> cp_tensor, dataset = simulated_random_cp_tensor((10, 20, 30), 3, noise_level=0.3, seed=0)
    >>> model_candidates = [
    ...     parafac(dataset, 3, init="random", random_state=0)
    ...     for i in range(5)
    ... ]

    Next, we sort the models by the error.

    >>> sorted_model_candidates, errors = sort_models_by_error(model_candidates, dataset)

    Now, the first element in sorted_model_candidates should be equal to the model with the lowest error.
    Let's double check by getting the model with the lowest error, and see which index it has.

    >>> lowest_error_model = get_model_with_lowest_error(model_candidates, dataset)
    >>> sorted_model_candidates.index(lowest_error_model)
    0

    Next, we can check if the errors are sorted

    >>> errors == sorted(errors)
    True
    """
    # TODOC: text example for sort_models_by_error
    errors = get_model_with_lowest_error(cp_tensors, dataset, error_function=error_function, return_errors=True)[1]
    sorted_errors = sorted(zip(errors, range(len(errors))))
    return (
        [cp_tensors[idx] for _error, idx in sorted_errors],
        [extract_singleton(error) for error, _cp_tensor in sorted_errors],
    )



# -*- coding: utf-8 -*-

__author__ = "Marie Roald & Yngve Mardal Moe"

from functools import wraps
from inspect import signature

import numpy as np
import pandas as pd
import xarray as xr


def is_iterable(x):
    """Check if variable is iterable

    Arguments
    ---------
    x
        Variable to check if is iterable

    Returns
    -------
    bool
        Whether ``x`` is iterable or not.
    """
    try:
        iter(x)
    except TypeError:
        return False
    else:
        return True


def is_xarray(x):
    """Check if ``x`` is an xarray DataArray.

    Arguments
    ---------
    x
        Object to check

    Returns
    -------
    bool
        ``True`` if x is an xarray DataArray, ``False`` otherwise.
    """
    return isinstance(x, xr.DataArray)


def is_dataframe(x):
    """Check if ``x`` is a DataFrame.

    Arguments
    ---------
    x
        Object to check

    Returns
    -------
    bool
        ``True`` if x is a DataFrame, ``False`` otherwise.
    """
    return isinstance(x, pd.DataFrame)


def _check_is_argument(func, arg_name):
    sig = signature(func)
    if arg_name in sig.parameters:
        return
    raise ValueError(f"{arg_name} is not an argument of {func}")


def _handle_none_weights_cp_tensor(cp_tensor_name, optional=False):
    def decorator(func):
        _check_is_argument(func, cp_tensor_name)

        @wraps(func)
        def func2(*args, **kwargs):
            bound_arguments = signature(func).bind(*args, **kwargs)

            if optional and cp_tensor_name not in bound_arguments.arguments:
                return func(*bound_arguments.args, **bound_arguments.kwargs)

            cp_tensor = bound_arguments.arguments.get(cp_tensor_name, None)
            weights, factors = validate_cp_tensor(cp_tensor)
            if weights is None:
                rank = factors[0].shape[1]
                cp_tensor = (np.ones(rank), factors)
                bound_arguments.arguments[cp_tensor_name] = cp_tensor

            out = func(*bound_arguments.args, **bound_arguments.kwargs)
            return out

        return func2

    return decorator


def validate_cp_tensor(cp_tensor):
    weights, factors = cp_tensor
    rank = factors[0].shape[1]
    for i, factor in enumerate(factors):
        if factor.ndim != 2:
            raise ValueError(
                f"All factor matrices should have two dimensions, but factor[{i}] has shape {factor.shape}."
            )

        # Check that all factor matrices have the same rank
        if factor.shape[1] != rank:
            raise ValueError(
                "All factor matrices should have the same number of columns. However factors[0] has shape"
                + f" {factors[0].shape} and factors[{i}] has shape {factor.shape}."
            )

    if weights is not None:
        if not isinstance(weights, np.ndarray):
            raise TypeError(f"The weights must be a numpy array, not {type(weights)}")
        if len(weights) != rank:
            raise ValueError("The weights should have the same length as the number of columns in the factor matrices.")
        if weights.ndim != 1:
            raise ValueError(f"The weights must be 1d array, {weights.ndim} != 1 ")
    return cp_tensor


_SINGLETON = object()
